import json
import time
import os
import threading
from typing import Union

# json format: https://docs.google.com/document/d/1CvAClvFfyA5R-PhYUmn5OOQtYMH4h6I0nSsKchNAySU/

__all__ = [
    "get_pid", "get_tid",
    "record_timestamp",
    "record_begin",
    "record_end",
    "record_duration",
    "record_thread_name",
    "record_process_name",
    "record_process_sort_index",
    "record_thread_sort_index",
    "record_dump",
    "record_init",
    "record_append",
    "record_begin_async",
    "record_end_async",
    "set_start_timestamp",
    "enable_trace",
    "disable_trace"
]


contents = []
START_TIMESTAMP = 0

TRACING_SWITCH=True

TRACING_STASH=True
TRACING_LOGGER=None

def enable_trace():
    global TRACING_SWITCH
    TRACING_SWITCH=True
def disable_trace():
    global TRACING_SWITCH
    TRACING_SWITCH=False

def should_trace(func):
    def inner(*args, **kwargs):
        global TRACING_SWITCH
        if TRACING_SWITCH:
            return func(*args, **kwargs)
        else:
            return
    return inner

def set_tracing_stash(tracing_stash=True, trace_file:str="trace.json", init_trace_file=True):
    import logging
    global TRACING_STASH, TRACING_LOGGER
    TRACING_STASH = tracing_stash
    if TRACING_STASH==False:
        # init TRACING_LOGGER
        TRACING_LOGGER = logging.getLogger("zq_tracing")
        TRACING_LOGGER.setLevel(logging.FATAL)
        file_handler = logging.FileHandler(trace_file)
        TRACING_LOGGER.addHandler(file_handler)
        if init_trace_file:
            with open(trace_file, 'w') as f:
                f.write("[\n")

    
def record_update(j:dict):
    if TRACING_STASH:
        contents.append(json.dumps(j))
    else:
        TRACING_LOGGER.fatal(json.dumps(j)+",")


def set_start_timestamp():
    global START_TIMESTAMP
    START_TIMESTAMP = time.time()


def get_pid():
    return os.getpid()

def get_tid():
    tid = threading.get_ident()
    if not tid: tid = 0
    return tid

@should_trace
def record_timestamp(name:str,
                 cat:str,
                 tid:int,
                 pid:int,
                 **kwargs) -> None:
    global START_TIMESTAMP
    j = {
        "name":name,
        "cat":cat,
        "ts": (time.time()-START_TIMESTAMP)*1000000,
        "pid": get_pid() if pid<0 else pid,
        "tid": get_tid() if tid<0 else tid
    }
    if kwargs:
        j["args"] = kwargs
    return j

@should_trace
def record_begin(name:str,
                 cat:str="",
                 tid=-1,
                 pid=-1,
                 **kwargs):
    j = record_timestamp(name, cat, tid, pid, **kwargs)
    j['ph'] = "B"
    record_update(j)

@should_trace
def record_end(name:str,
               cat:str="",
               tid=-1,
               pid=-1,
               **kwargs):
    j = record_timestamp(name, cat, tid, pid, **kwargs)
    j['ph'] = "E"
    record_update(j)
    
@should_trace
def record_begin_async(name:str,
                       id:Union[str,int],
                       cat:str="",
                       tid=-1,
                       pid=-1,
                       **kwargs
                       ):
    j = record_timestamp(name, cat, tid, pid, **kwargs)
    j['ph'] = 'b'
    j['id'] = id
    record_update(j)
    
@should_trace
def record_end_async(name:str,
                       id:Union[str,int],
                       cat:str="",
                       tid=-1,
                       pid=-1,
                       **kwargs
                       ):
    j = record_timestamp(name, cat, tid, pid, **kwargs)
    j['ph'] = 'e'
    j['id'] = id
    record_update(j)

@should_trace
def record_duration(name:str,
                    cat:str="",
                    tid=-1,
                    pid=-1,
                    dur:float=0,
                    **kwargs):
    j = record_timestamp(name, cat, tid, pid, **kwargs)
    j['ph'] = "X"
    j['dur'] = dur  
    record_update(j)
    
@should_trace
def record_thread_name(name:str, tid=-1, pid=-1, **kwargs):
    j = {
        "name": "thread_name",
        "ph": "M",
        "pid": get_pid() if pid<0 else pid,
        "tid": get_tid() if tid<0 else tid,
    }
    kwargs['name'] = name
    j['args'] = kwargs
    record_update(j)

@should_trace
def record_process_name(name:str, pid=-1, **kwargs):
    j = {
        "name": "process_name",
        'ph': 'M',
        'pid': get_pid() if pid<0 else pid,
    }
    kwargs['name'] = name
    j['args'] = kwargs
    record_update(j)
    
@should_trace
def record_process_sort_index(index:int, pid=-1, **kwargs):
    j = {
        'name': 'process_sort_index',
        'ph': 'M',
        'pid': get_pid() if pid<0 else pid,
    }
    kwargs['sort_index'] = index
    j['args'] = kwargs
    record_update(j)

@should_trace
def record_thread_sort_index(index:int, tid=-1, pid=-1, **kwargs):
    j = {
        'name': 'thread_sort_index',
        'ph': 'M',
        'pid': get_pid() if pid<0 else pid,
        'tid': get_tid() if tid<0 else tid,
    }
    kwargs['sort_index'] = index
    j['args'] = kwargs
    record_update(j)

@should_trace
def record_dump(filename:str):
    with open(filename, 'w') as f:
        f.write("[\n")
        f.write(",\n".join(contents))
        f.write("\n]\n")

@should_trace
def record_init(filename:str):
    with open(filename, 'w') as f:
        f.write("[\n")

@should_trace
def record_append(filename:str):
    with open(filename, 'a') as f:
        for content in contents:
            f.write(f"{content},\n")

        
         
        
if __name__ == '__main__':
    # disable_trace()
    def demo1():
        for i in range(2):
            print(f"{i}")
            record_begin(name=f"name_{i}", cat="")
            time.sleep(1)
            record_end(name=f"name_{i}", cat="")
            time.sleep(0.5)
        record_thread_name(name="thread")
        record_process_name(name="process")
        record_dump("./test.json")
    
    def demo2():
        set_tracing_stash(False, trace_file="demo2.json", init_trace_file=True)
        # with open("demo2.json", 'w') as f: f.write("[\n")
        for i in range(2):
            print(f"{i}")
            record_begin(name=f"name_{i}", cat="")
            time.sleep(1)
            record_end(name=f"name_{i}", cat="")
            time.sleep(0.5)
        record_thread_name(name="thread")
        record_process_name(name="process")
    
    demo2()


import os
from typing import Generator

def yield_files(path:str, recursive=False)-> Generator[None, None, str]:
    if recursive==False:
        for v in os.listdir(path):
            yield os.path.join(path, v)
    else:
        for v in os.listdir(path):
            pv = os.path.join(path, v)
            if os.path.isdir(pv):
                yield from yield_files(pv, recursive)
            else:
                yield pv

if __name__ == '__main__':
    for v in yield_files("..", recursive=True):
        print(v)


import time
import sys
import functools
import inspect
from typing import Any, Callable, TypeVar, cast
from zq_tools.zq_logger import default_logger as logger

# Used for annotating the decorator usage of 'no_grad' and 'enable_grad'.
# See https://mypy.readthedocs.io/en/latest/generics.html#declaring-decorators
FuncType = Callable[..., Any]
F = TypeVar('F', bound=FuncType)


class _DecoratorContextManager:
    """Allow a context manager to be used as a decorator"""
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
    def __call__(self, func: F) -> F:
        if inspect.isgeneratorfunction(func):
            return self._wrap_generator(func)

        @functools.wraps(func)
        def decorate_context(*args, **kwargs):
            with self.clone():
                return func(*args, **kwargs)
        return cast(F, decorate_context)

    def _wrap_generator(self, func):
        """Wrap each generator invocation with the context manager"""
        @functools.wraps(func)
        def generator_context(*args, **kwargs):
            gen = func(*args, **kwargs)

            # Generators are suspended and unsuspended at `yield`, hence we
            # make sure the grad mode is properly set every time the execution
            # flow returns into the wrapped generator and restored when it
            # returns through our `yield` to our caller (see PR #49017).
            try:
                # Issuing `None` to a generator fires it up
                with self.clone():
                    response = gen.send(None)

                while True:
                    try:
                        # Forward the response to our caller and get its next request
                        request = yield response

                    except GeneratorExit:
                        # Inform the still active generator about its imminent closure
                        with self.clone():
                            gen.close()
                        raise

                    except BaseException:
                        # Propagate the exception thrown at us by the caller
                        with self.clone():
                            response = gen.throw(*sys.exc_info())

                    else:
                        # Pass the last request to the generator and get its response
                        with self.clone():
                            response = gen.send(request)

            # We let the exceptions raised above by the generator's `.throw` or
            # `.send` methods bubble up to our caller, except for StopIteration
            except StopIteration as e:
                # The generator informed us that it is done: take whatever its
                # returned value (if any) was and indicate that we're done too
                # by returning it (see docs for python's return-statement).
                return e.value

        return generator_context

    def __enter__(self) -> None:
        raise NotImplementedError

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        raise NotImplementedError

    def clone(self):
        # override this method if your children class takes __init__ parameters
        return self.__class__(*self.args, **self.kwargs)

def do_nothing(*args, **kwargs): 
    pass

def pass_it(func): # decorator
    return do_nothing

class time_it(_DecoratorContextManager):
    def __init__(self, 
                 keyword="", 
                 print_it=print, 
                 sync_func=do_nothing):
        super().__init__(keyword=keyword, print_it=print_it, sync_func=sync_func)
        self.print_it = print_it
        self.sync_func = sync_func
        self.keyword = keyword if keyword=="" else f"[{keyword}] "
    def __enter__(self):
        self.sync_func()
        self.start_time = time.time()
    def __exit__(self, exc_type: Any, exc_value: Any, traceback:Any):
        self.sync_func()
        self.stop_time = time.time()
        self.print_it(f"{self.keyword}time_it: {self.stop_time-self.start_time}s")

    
if __name__ == '__main__':
    @pass_it
    def test():
        print("hello in test")
    test()
    
    @time_it(keyword="time it as wrapper")
    def test():
        print("hello in decorator")
    test()
    
    with time_it(keyword="time it as context manager"):
        print("hello in `with`")

    from zq_tools.zq_logger import default_logger as logger
    with time_it(keyword="time it as context manager", print_it=lambda x:logger.info(x)):
        print("hello in `with`, print with logger")
    
    
    


import logging
import colorful as cf
import os

__all__ = ["get_logger", "default_logger"]
allocated_loggers = {}

class ZQ_Logger(logging.Logger):
    DEBUG=logging.DEBUG
    INFO=logging.INFO
    WARN=logging.WARN
    WARNING=logging.WARNING
    FATAL=logging.FATAL
    CRITICAL=logging.CRITICAL
    PRANK = 999
    
    def default_color(self, x): return x
    color_to_rank = [
        cf.italic_red,
        cf.italic_yellow,
        cf.italic_cyan,
        cf.italic_orange,
        cf.italic_blue,
        cf.italic_magenta,
        cf.italic_green,
        cf.italic_purple,
    ]
    def __init__(self, name):
        super(ZQ_Logger, self).__init__(name)
        self.tag = ""
        self.print_thread = False
        self.print_level = True
        self.rank = 0
        self.name2handler = dict()
        logging.Logger.setLevel(self, logging.DEBUG)

        
    def add_log_file(self, log_file:str, name:str=""):
        if not name: name = log_file
        if name in self.name2handler: return
        handler = logging.FileHandler(log_file)
        self.name2handler[name] = handler
        self.addHandler(handler)
        self.reset_format()

    def set_level_for_handler(self, name:str, level:int):
        if name not in self.name2handler: return
        handler: logging.Handler = self.name2handler[name]
        handler.setLevel(level)
        
    def set_level_for_all(self, level:int):
        for name in self.name2handler:
            handler: logging.Handler = self.name2handler[name]
            handler.setLevel(level)
    
    def setLevel(self, *args, **kwargs):
        print(f"Warn: `setLevel` is not supported, use `set_level_for_all` instead")
        
        
        
    def generate_fmt(self)->logging.StreamHandler:
        thread_fmt = "" if not self.print_thread else "[%(threadName)s] "
        level_fmt = "" if not self.print_level else " [%(levelname)s]"
        basic_fmt = f'[%(asctime)s.%(msecs)03d] {thread_fmt}"%(pathname)s", line %(lineno)d{level_fmt}:{self.tag} %(message)s'
        date_fmt = "%Y-%m-%d %H:%M:%S"
        fmt = logging.Formatter(
            fmt = basic_fmt,
            datefmt = date_fmt
        )
        return fmt
    
    def set_rank(self, rank:int):
        self.rank = rank
        self._set_tag(f"[Rank {rank}]")
        self.default_color = self.color_to_rank[rank%len(self.color_to_rank)]
        return self
        
    def reset_format(self):
        formatter = self.generate_fmt()
        for handler in self.handlers: 
            handler.setFormatter(formatter)
        return self
            
    def _set_tag(self, tag:str):
        self.tag = tag
        self.reset_format()
        return self

    def set_print_thread(self, print_thread:bool=True):
        self.print_thread = print_thread
        self.reset_format()
        return self
    
    def prank(self, msg:str, color:str='',*args, **kwargs):
        '''print with rank. If color is not specified, use the color format corresponding to the rank'''
        if not self.isEnabledFor(self.PRANK): return
        color = getattr(cf, color) if color else self.default_color
        self._log(self.PRANK, color(msg), args, **kwargs)
    # def debug(self, msg:str, color:str='',*args, **kwargs):
    #     '''print with rank. If color is not specified, use the color format corresponding to the rank'''
    #     if not self.isEnabledFor(self.DEBUG): return
    #     color = getattr(cf, color) if color else self.default_color
    #     self._log(self.DEBUG, color(msg), args, **kwargs)
    def debug(self, msg:str, *args, **kwargs):
        if self.isEnabledFor(logging.INFO): self._log(logging.DEBUG, msg, args, **kwargs)
    def info(self, msg:str, *args, **kwargs):
        if self.isEnabledFor(logging.INFO): self._log(logging.INFO, cf.green(msg), args, **kwargs)
    def warn(self, msg:str, *args, **kwargs):
        if self.isEnabledFor(logging.WARN): self._log(logging.WARN, cf.yellow(msg), args, **kwargs)
    def error(self, msg:str, *args, **kwargs):
        if self.isEnabledFor(logging.ERROR): self._log(logging.ERROR, cf.red(msg), args, **kwargs)
    def fatal(self, msg:str, *args, **kwargs):
        if self.isEnabledFor(logging.FATAL): self._log(logging.FATAL, cf.bold_red(msg), args, **kwargs)

    def prank_root(self, msg:str, color:str='', root=0, *args, **kwargs):
        '''print with rank. If color is not specified, use the color format corresponding to the rank'''
        if self.rank != root: return
        if not self.isEnabledFor(self.PRANK): return
        color = getattr(cf, color) if color else self.default_color
        self._log(self.PRANK, color(msg), args, **kwargs)
    def debug_root(self, msg:str, color:str='', root=0, *args, **kwargs):
        '''print with rank. If color is not specified, use the color format corresponding to the rank'''
        if self.rank != root: return
        if not self.isEnabledFor(self.DEBUG): return
        color = getattr(cf, color) if color else self.default_color
        self._log(self.DEBUG, color(msg), args, **kwargs)
    def info_root(self, msg:str, root=0, *args, **kwargs):
        if self.rank != root: return
        if self.isEnabledFor(logging.INFO): self._log(logging.INFO, cf.green(msg), args, **kwargs)
    def warn_root(self, msg:str, root=0, *args, **kwargs):
        if self.rank != root: return
        if self.isEnabledFor(logging.WARN): self._log(logging.WARN, cf.yellow(msg), args, **kwargs)
    def error_root(self, msg:str, root=0, *args, **kwargs):
        if self.rank != root: return
        if self.isEnabledFor(logging.ERROR): self._log(logging.ERROR, cf.red(msg), args, **kwargs)
    def fatal_root(self, msg:str, root=0, *args, **kwargs):
        if self.rank != root: return
        if self.isEnabledFor(logging.FATAL): self._log(logging.FATAL, cf.bold_red(msg), args, **kwargs)
        
    warning = warn
    critical = fatal
    warning_root = warn_root
    critical_root = fatal_root
    
def get_level_from_env(logger_name:str, default_level="info"):
    level = default_level if logger_name not in os.environ else os.environ[logger_name]
    level = level.lower()
    level2num = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warn": logging.WARN,
        "warning": logging.WARN,
        "error": logging.ERROR,
        "fatal": logging.FATAL,
        "critical": logging.FATAL,
    }
    if level in level2num: return level2num[level]
    print(f"Unknown level {level} for logger {logger_name}, use default level {default_level}")
    return level2num[default_level]

    

def get_logger(logger_name="Z_LEVEL",
               enable_console = True)->ZQ_Logger:
    if logger_name in allocated_loggers: return allocated_loggers[logger_name]
    # why need to call `setLoggerClass` twice? refer to the issue: https://bugs.python.org/issue37258
    logging.setLoggerClass(ZQ_Logger)
    logger:ZQ_Logger = logging.getLogger(logger_name)
    logging.setLoggerClass(logging.Logger)
    # Initilize level from environment. If not specified, use INFO
    if enable_console:
        streamHandler = logging.StreamHandler()
        name = logger_name
        logger.name2handler[name] = streamHandler
        streamHandler.setLevel(get_level_from_env(logger_name))
        logger.addHandler(streamHandler)
    logger.reset_format()
    allocated_loggers[logger_name] = logger
    return logger

default_logger = get_logger()


if __name__ == '__main__':
    def test_environ():
        print(f'{"="*20} test environ {"="*20}')
        logger1 = get_logger("logger1")
        logger1.debug("this message should not be printed due to default initilizing level is INFO")
        logger1.set_level_for_all(logger1.DEBUG)
        logger1.debug("this message should be printed due to call `set_level_for_all`")
        os.environ['logger2'] = "debug"
        logger2 = get_logger("logger2")
        logger2.debug("this message should be printed due to env `logger` is set to `debug`")
    
    def test_log_file():
        print(f'{"="*20} test log file {"="*20}')
        # recommend to use `ANSI Color` extention to view log file in VSCode
        logger = get_logger("test_log_file")
        logger.add_log_file("demo.log")
        logger.info(f"this message should be printed to both console and file {logger.name2handler.keys()}")
        
    def test_ranks():
        print(f'{"="*20} test ranks {"="*20}')
        logger=get_logger("test_ranks")
        logger.set_level_for_handler("test_ranks", logger.DEBUG)
        logger.debug_root("printed due to default rank is 0 and default style is plain")
        logger.debug_root("NOT printed due to default rank is 0", root=2)
        for rank in range(8):
            logger.set_rank(rank)
            logger.debug(f"style of rank {rank}")

    def test_styles():
        print(f'{"="*20} test styles {"="*20}')
        logger=get_logger("test_styles")
        logger.info("style of info msg (green)")
        logger.warn("style of warn msg (yellow)")
        logger.error("style of error msg (red) ")
        logger.fatal("style of fatal msg (bold red)")
    
    def test_threads():
        print(f'{"="*20} test threads {"="*20}')
        logger = get_logger("test_threads")
        logger.set_print_thread(print_thread=True)
        logger.info(f"this message should be printed with thread id")
        

    test_environ()
    test_log_file()
    test_ranks()
    test_styles()
    test_threads()
    
    